import inspect
from typing import get_type_hints, get_origin, get_args, Callable, Any, Literal
from enum import Enum

def _python_type_to_json_type(py_type: Any) -> dict:
    origin = get_origin(py_type)
    args = get_args(py_type)

    # Handle Literal
    if origin is Literal:
        return {
            "type": _python_type_to_json_type(type(args[0]))["type"],
            "enum": list(args)
        }

    # Handle Enum
    if isinstance(py_type, type) and issubclass(py_type, Enum):
        return {
            "type": "string",
            "enum": [e.value for e in py_type]
        }

    # Handle List[X]
    if origin == list and args:
        return {
            "type": "array",
            "items": _python_type_to_json_type(args[0])
        }

    # Basic type mapping
    mapping = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object"
    }

    return {"type": mapping.get(origin or py_type, "string")}


def _parse_pydoc(func: Callable) -> dict:
    doc = inspect.getdoc(func) or ""
    lines = doc.splitlines()
    description_lines = []
    param_descriptions = {}

    for line in lines:
        stripped = line.strip()
        # Coleta descrição geral até encontrar o primeiro ':param'
        if stripped.startswith(":param"):
            break
        description_lines.append(stripped)

    for line in lines:
        line = line.strip()
        if line.startswith(":param"):
            try:
                param, desc = line[7:].split(":", 1)
                param_descriptions[param.strip()] = desc.strip()
            except ValueError:
                continue  # Ignora linha mal formatada

    description = " ".join(description_lines).strip()

    return {
        "description": description,
        "param_descriptions": param_descriptions
    }


def to_openai_tool(func: Callable, new_format: bool) -> dict:
    sig = inspect.signature(func)
    hints = get_type_hints(func)
    doc_info = _parse_pydoc(func)

    properties = {}
    for name, param in sig.parameters.items():
        param_type = hints.get(name, str)
        json_schema = _python_type_to_json_type(param_type)
        json_schema["description"] = doc_info["param_descriptions"].get(name, "No description provided.")
        properties[name] = json_schema

    if new_format:
        return {
            "type": "function",
            "name": func.__name__,
            "description": doc_info["description"],
            "parameters": {
                "type": "object",
                "properties": properties,
                "required": list(sig.parameters),
            },
        }
    else:
        return {
            "type": "function",
            "function": {
                "name": func.__name__,
                "description": doc_info["description"],
                "parameters": {
                    "type": "object",
                    "properties": properties,
                    "required": list(sig.parameters),
                    "additionalProperties": False
                },
                "strict": True
            }
        }



import json
import asyncio
from typing import List, Dict, Tuple, Optional, Callable, Union
from datetime import datetime
from tyr_agent.entities.entities import ManagerCallManyAgents, AgentCallInfo, AgentHistory, AgentInteraction
from tyr_agent.models.gemini_model import GeminiModel
from tyr_agent.models.gpt_model import GPTModel
from tyr_agent.storage.interaction_history import InteractionHistory
import uuid


class SimpleAgent:
    MAX_ALLOWED_HISTORY = 20

    def __init__(self, prompt_build: str, agent_name: str, model: Union[GeminiModel, GPTModel], storage: Optional[InteractionHistory] = None, max_history: int = 20, use_storage: bool = True, use_history: bool = True, use_score: bool = True, score_average: Union[int, float] = 3):
        self.prompt_build: str = prompt_build
        self.agent_name: str = agent_name

        self.storage: Optional[InteractionHistory] = None
        self.history: Optional[List[dict]] = None
        self.use_storage: bool = use_storage
        self.use_history: bool = use_history

        self.use_score: bool = use_score
        self.score_average: Union[int, float] = score_average if self._is_valid_score(score_average) else 3

        if use_storage and use_history:
            self.storage = storage or InteractionHistory(f"{agent_name.lower()}_history.json")
            self.history = self.storage.load_history(agent_name)

            if use_score:
                self._filter_history_by_score()

        self.agent_model: Union[GeminiModel, GPTModel] = model

        self.MAX_HISTORY = min(max_history, self.MAX_ALLOWED_HISTORY)
        self.PROMPT_TEMPLATE = ""

    async def chat(self, user_input: str, streaming: bool = False, files: Optional[List[dict]] = None, save_history: bool = True) -> Optional[str]:
        try:
            agent_response: str = self.agent_model.generate(self.prompt_build, user_input, files, self.history, self.use_history)

            if (self.use_history or self.use_storage) and save_history:
                self._update_history(user_input, [agent_response], "simple")

            return agent_response
        except Exception as e:
            print(f"❌ [SimpleAgent.chat] {type(e).__name__}: {e}")
            return None

    def _update_history(self, user_input: str, agent_response: List[str], type_agent: str, called_functions: List[dict] | None = None, score: int | None = None) -> None:
        try:
            actual_conversation = {
                "id": str(uuid.uuid4()),
                "timestamp": datetime.now().strftime("%Y-%m-%dT%H:%M:%S"),
                "interaction": {
                    "user": user_input,
                    "agent": agent_response
                },
                "called_functions": called_functions if called_functions is not None else [],
                "type_agent": type_agent,
                "score": score
            }

            if self.history and self.use_history:
                self.history.append(actual_conversation)
                self.history = self.history[-self.MAX_HISTORY:]  # -> Mantendo apenas os N itens no histórico.

            if self.storage and self.use_storage:
                self.storage.save_history(self.agent_name, actual_conversation)
        except Exception as e:
            print(f'[ERROR] - Ocorreu um erro duração a atualização do histórico: {e}')

    def get_agent_history(self) -> List[dict]:
        return self.history

    def create_agent_history_with_storage(self, storage: Optional[InteractionHistory] = None, use_history: bool = True) -> None:
        """
        Cria uma instância de histórico para o agente a partir de um storage.
        Caso já exista um histórico para o agente, apenas conecta o histórico com o agente.
        :param storage: Histórico (Storage) a ser carregado, caso não seja passado será criado/procurado um.
        :param use_history: Define qual o novo valor de use_history do agente, por padrão é True.
        :return: Não retorna nada.
        """
        self.storage: InteractionHistory | None = storage or InteractionHistory(f"{self.agent_name.lower()}_history.json")
        self.history: List[dict] | None = self.storage.load_history(self.agent_name)
        self.use_history: bool = use_history

    def create_agent_history(self, new_history: List[AgentHistory], use_history: bool = True) -> bool:
        """
        Cria um histórico para o agente a patir de uma lista de AgentHistory.
        Caso já exista um histórico para o agente, reescreve o histórico atual.
        Altera o valor use_history para true automaticamente.
        :param new_history: Histórico (History) a ser carregado.
        :param use_history: Define qual o novo valor de use_history do agente, por padrão é True.
        :return: Retorna true caso tudo tenha dado certo e false caso tenha acontecido algum problema.
        """
        try:
            self.history: List[dict] = self._format_history(new_history)
            self.use_history: bool = use_history
            return True
        except Exception as e:
            return False

    def extend_agent_history(self, extended_history: List[AgentHistory]) -> bool:
        """
        Insere novos registros no histórico do agente através de uma lista de AgentHistory.
        Apenas insere os novos registro se o agente tiver um histórico.
        :param extended_history: Histórico (History) a ser carregado.
        :return: Retorna true caso tudo tenha dado certo e false caso tenha acontecido algum problema.
        """
        try:
            if self.history:
                self.history.extend(self._format_history(extended_history))
                return True
            else:
                return False
        except Exception as e:
            return False

    def remove_agent_history(self, use_history: bool = False) -> None:
        """
        Remove o histórico carregado da instância atual.
        Não exclui o arquivo físico do histórico no disco.
        :param use_history: Decide se o histórico vai continuar sendo usado, por padrão é False.
        :return: None
        """
        self.storage: InteractionHistory | None = None
        self.history: List[dict] | None = None
        self.use_history: bool = use_history

    def clear_agent_history(self) -> None:
        """
        Limpa o campo histórico da agente, caso ele exista.
        Não altera o arquivo do histórico no disco.
        :return: None
        """
        if self.history is not None:
            self.history.clear()

    def clear_agent_storage(self) -> None:
        """
        Limpa o campo storage do agente, caso ele exista.
        Limpa o arquivo do histórico no disco.
        :return: None
        """
        if self.storage is not None:
            self.storage.clear_history()

    def rate_interaction(self, interaction_id: str, score: Union[int, float]) -> Tuple[bool, bool]:
        """
        Define o score de uma interação específica do histórico.
        Pode ser feita apenas no history ou através do storage, caso ele esteja sendo usado.
        Atualiza o histórico atual do agente baseado no score_average do agente.
        :param interaction_id: ID da interação.
        :param score: Nota definida para a interação, indo apenas de 0 a 5.
        :return: Retorna uma tupla na seguinte lógica: (response_update_history, response_update_storage)
        """
        try:
            if not self.history and not self.storage:
                return False, False

            if not self._is_valid_score(score):
                return False, False

            response_update_history: bool = False
            response_update_storage: bool = True

            if self.history:
                # Atualizando o history:
                for interaction in self.history:
                    if interaction["id"] == interaction_id:
                        interaction["score"] = score
                        break
                response_update_history = True

            if self.storage:
                # Atualizando o storage:
                response_update_storage = self.storage.update_score(self.agent_name, interaction_id, score)

            if self.use_score:
                # Filtrando o "novo" histórico. (Uma interação foi avaliada, então preciso verificar se ela vai sair)
                if not self._is_valid_score(self.score_average):
                    self.score_average = 3
                self._filter_history_by_score()

            return response_update_history, response_update_storage
        except Exception as e:
            print(e)
            return False, False

    def delete_interaction(self, interaction_id: str) -> Tuple[bool, bool]:
        """
        Deleta a interação com o id informado.
        Caso o agente não utilize o storage, deleta apenas a interação do history.
        :param interaction_id: ID da interação que será deletada.
        :return: Retorna uma tupla na seguinte lógica: (response_delete_from_history, response_delete_from_storage)
        """
        try:
            if not self.history and not self.storage:
                return False, False

            response_delete_from_history: bool = False
            response_delete_from_storage: bool = True

            if self.history:
                self.history = list(filter(lambda x: x["id"] != interaction_id, self.history))
                response_delete_from_history = True

            if self.storage:
                response_delete_from_storage = self.storage.delete_history(self.agent_name, interaction_id)

            return response_delete_from_history, response_delete_from_storage
        except Exception as e:
            return False, False

    def get_score_by_id(self, interaction_id: str, find_by_history: bool) -> Union[int, float]:
        """
        Pega o score da interação procurada.
        Pode ser feita apenas no history ou através do storage, caso ele esteja sendo usado.
        :param interaction_id: Id da interação procurada.
        :param find_by_history: Define se a busca será feita no history ou no storage.
        :return: Retorna o score de uma interação.
        """
        try:
            if not self.history and not self.storage:
                return 0.0

            if find_by_history and self.history:
                # Procurando no history, se ele existir:
                for interaction in self.history:
                    if interaction.get("id") == interaction_id:
                        return interaction.get("score")

            if not find_by_history and self.storage:
                # Procurando no storage, se ele existir:
                data = self.storage.load_all()

                for interaction in data.get(self.agent_name, []):
                    if interaction.get("id") == interaction_id:
                        return interaction.get("score")

            return 0.0
        except Exception as e:
            print(f"[ERROR] - get_score_by_id: {e}")
            return 0.0

    def get_average_score(self, find_by_history: bool) -> float:
        """
        Pega a média do score no histórico.
        :param find_by_history: Define se a busca será feita no history ou no storage.
        :return: Média do score.
        """
        try:
            if not self.history and not self.storage:
                return 0.0

            sum_interactions: Union[int, float] = 0
            count_interactions: int = 0

            if find_by_history and self.history:
                for interaction in self.history:
                    if isinstance(interaction.get("score"), (int, float)):
                        sum_interactions += interaction.get("score")
                        count_interactions += 1

                if count_interactions == 0:
                    return 0.0

                return sum_interactions / count_interactions

            if not find_by_history and self.storage:
                data = self.storage.load_all()

                for interaction in data.get(self.agent_name, []):
                    if isinstance(interaction.get("score"), (int, float)):
                        sum_interactions += interaction.get("score")
                        count_interactions += 1

                if count_interactions == 0:
                    return 0.0

                return sum_interactions / count_interactions

            return 0.0
        except Exception as e:
            return 0.0

    def get_all_scores(self, find_by_history: bool) -> List[dict]:
        """
        Pega todos os scores de um agente.
        :param find_by_history: Define se a busca será feita no history ou no storage.
        :return: Retorna uma lista de dicionários contendo todos os ids e scores.
        """
        try:
            if not self.history and not self.storage:
                return []

            if find_by_history and self.history:
                return [
                    {"id": i.get("id"), "score": i.get("score")}
                    for i in self.history
                    if "id" in i
                ]

            if not find_by_history and self.storage:
                data = self.storage.load_all()

                return [
                    {"id": i.get("id"), "score": i.get("score")}
                    for i in data.get(self.agent_name, [])
                    if "id" in i
                ]

            return []
        except Exception as e:
            return []

    def _is_valid_score(self, score: Union[int, float]) -> bool:
        """
        Verifica se o score informado é válido.
        :param score: Score a ser validado.
        :return: True caso seja válido. | False caso não seja válido.
        """
        return isinstance(score, (int, float)) and 0 <= score <= 5

    def _filter_history_by_score(self) -> None:
        """
        Filtra o histórico atual com base no score_average do agente.
        :return: None
        """
        self.history = list(filter(lambda x: (x.get("score") is None or isinstance(x.get("score"), (int, float))) and (x.get("score") is None or x.get("score") >= self.score_average), self.history))

    def _format_history(self, target_history: List[AgentHistory]) -> List[dict]:
        temp_history: List[dict] = []
        for h in target_history:
            if h.get("timestamp") is None:
                timestamp: str = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            else:
                timestamp: str = h.get("timestamp")

            temp_history.append(
                {
                    "id": str(uuid.uuid4()),
                    "timestamp": timestamp,
                    "interaction": {
                        "user": h["interaction"]["user"],
                        "agent": h["interaction"]["agent"]
                    },
                    "called_functions": [],
                    "type_agent": h["type_agent"],
                    "score": h.get("score") if h.get("score", False) else None,
                }
            )
        return temp_history


class ComplexAgent(SimpleAgent):
    MAX_ALLOWED_HISTORY = 20

    def __init__(self, prompt_build: str, agent_name: str, model: Union[GeminiModel, GPTModel], functions: Optional[List[Callable]] = None, final_prompt: Optional[str] = None, storage: Optional[InteractionHistory] = None, max_history: int = 20, use_storage: bool = True, use_history: bool = True, use_score: bool = True, score_average: Union[int, float] = 3):
        super().__init__(prompt_build, agent_name, model, storage, max_history, use_storage, use_history, use_score, score_average)
        self.functions: Optional[List[Callable]] = functions or {}

        self.final_prompt = final_prompt

    async def chat(self, user_input: str, streaming: bool = False, files: Optional[List[dict]] = None, save_history: bool = True) -> Optional[str]:
        try:
            agent_response = await self.agent_model.generate_with_functions(self.prompt_build, user_input, files, self.history, self.use_history, self.functions, self.final_prompt)

            if (self.use_history or self.use_storage) and save_history:
                self._update_history(user_input, [agent_response], "complex")

            return agent_response
        except Exception as e:
            print(f'[ERROR] - Ocorreu um erro durante a comunicação com o agente: {e}')
            return None


class ManagerAgent(SimpleAgent):
    MAX_ALLOWED_HISTORY = 100

    def __init__(self, agent_name: str, model: Union[GeminiModel, GPTModel], agents: List[Union[SimpleAgent, ComplexAgent]], storage: Optional[InteractionHistory] = None, max_history: int = 100, use_storage: bool = True, use_history: bool = True, use_score: bool = True, score_average: Union[int, float] = 3):
        super().__init__("", agent_name, model, storage, max_history, use_storage, use_history, use_score, score_average)

        self.agents: Dict[str, Union[SimpleAgent, ComplexAgent]] = {agent.agent_name: agent for agent in agents}

    async def chat(self, user_input: str, streaming: bool = False, files: Optional[List[dict]] = None, save_history: bool = True) -> Optional[str]:
        # Gera o prompt com base nos agentes disponíveis:
        prompt: str = self.__generate_prompt()

        if not prompt:
            print(f"[ERRO] Não foi possível montar o prompt.")
            return None

        try:
            agent_response: str = self.agent_model.generate(prompt, user_input, None, self.history, self.use_history)

            extracted_agents = self.__extract_agent_call(agent_response)

            if not extracted_agents:
                if self.use_history and save_history:
                    self.__update_history(user_input, agent_response, False, [])
                return agent_response

            # Encontrando os Agentes solicitados:
            delegated_agents = self.__find_correct_agents(extracted_agents)

            if len(delegated_agents) == 0:
                requested_agents: str = extracted_agents if isinstance(extracted_agents, str) else json.dumps(extracted_agents, ensure_ascii=False)
                print(f"[ERRO] Nenhum dos agentes requisitados foi encontrado: {requested_agents}")
                return None

            response_delegated_agents = await self.__execute_agents_calls(delegated_agents)

            final_prompt: str = self.__generate_final_prompt(response_delegated_agents)

            if not final_prompt:
                return "\n".join(f"{k}: {v}" for agent in response_delegated_agents for k, v in agent.items())

            final_agent_response: str = self.agent_model.generate(final_prompt, user_input, None, None, False)

            if (self.use_history or self.use_storage) and save_history:
                self.__update_history(user_input, agent_response, True, response_delegated_agents)

            return final_agent_response

        except Exception as e:
            print(f"[ERROR] - Falha ao interpretar a resposta do manager: {e}")
            return None

    def __extract_agent_call(self, response_text: str) -> Optional[ManagerCallManyAgents]:
        try:
            text_cleaned = (
                response_text.removeprefix("```json\n").removesuffix("\n```").replace("\n", "")
                .replace("`", "").replace("´", "").strip()
            )

            data = json.loads(text_cleaned)
            if isinstance(data, dict) and "call_agents" in data and "agents_to_call" in data:
                return data
            return None
        except json.JSONDecodeError:
            return None

    def __find_correct_agents(self, agents_to_call: ManagerCallManyAgents) -> List[AgentCallInfo]:
        try:
            agents = []
            for agent in agents_to_call.get("agents_to_call", []):
                if agent.get("agent_to_call", "") not in self.agents.keys():
                    raise Exception("Erro ao procurar o agente correspondente.")
                else:
                    agents.append({"agent": self.agents[agent.get("agent_to_call")], "message": agent.get("agent_message")})

            # Encontrando o Agente solicitado:
            return agents
        except Exception as e:
            print(f"[ERROR] - Falha ao encontrar o agente responsável: {e}")
            return []

    async def __execute_agents_calls(self, delegated_agents: List[AgentCallInfo]) -> List[dict]:
        # Execução paralela dos agentes:
        coroutines = [
            delegated_agent["agent"].chat(delegated_agent["message"], streaming=True)
            for delegated_agent in delegated_agents
        ]

        results = await asyncio.gather(*coroutines, return_exceptions=True)

        agents_response: List[dict] = []

        for agent_info, result in zip(delegated_agents, results):
            agent_name = agent_info["agent"].agent_name

            if isinstance(result, Exception):
                print(f"[ERRO] Agente '{agent_name}' falhou: {type(result).__name__} - {result}")
                agents_response.append({agent_name: "[Erro ao gerar resposta]"})
            else:
                if isinstance(result, str):
                    agents_response.append({agent_name: result})

        return agents_response

    def __generate_prompt(self) -> str:
        try:
            formatted_agents = "\n".join(
                f"- {agent_name}: {agent.prompt_build}\n" for agent_name, agent in
                self.agents.items())

            first_prompt_template: str = """Você é um agente gerente responsável por delegar perguntas aos agentes adequados.

INSTRUÇÕES OBRIGATÓRIAS:
- Responda EXCLUSIVAMENTE com JSON puro ao delegar a agentes.
- Nunca misture JSON com texto comum, markdown ou comentários.
- Se nenhum agente puder responder, responda com texto comum (sem JSON).
- Agrupe perguntas destinadas ao mesmo agente em uma única mensagem.

FORMATO DO JSON:
{
  "call_agents": true,
  "agents_to_call": [
    {"agent_to_call": "nome_do_agente", "agent_message": "mensagem_concatenada"}
  ]
}
"""

            second_prompt_template: str = f"""
AGENTES DISPONÍVEIS:

{formatted_agents}"""

            third_prompt_template: str = """
EXEMPLOS:

Chamada única:
{"call_agents": true, "agents_to_call": [{"agent_to_call": "MathAgent", "agent_message": "Quanto é 10+5?"}]}

Chamada múltipla:
{"call_agents": true, "agents_to_call": [{"agent_to_call": "MathAgent", "agent_message": "Quanto é 20+20?"}, {"agent_to_call": "WeatherAgent", "agent_message": "Qual é o clima no Rio?"}]}

Agrupamento para o mesmo agente:
{"call_agents": true, "agents_to_call": [{"agent_to_call": "MathAgent", "agent_message": "Quanto é 5+5? Quanto é 10+10?"}]}

Resposta direta (sem agentes):
Claro! Posso ajudar diretamente com essa questão."""

            return first_prompt_template + second_prompt_template + third_prompt_template
        except Exception as e:
            print(f'[ERROR] - Ocorreu um erro durante a geração do prompt: {e}')
            return ""

    def __generate_final_prompt(self, agents_response: List[dict]) -> str:
        try:
            combined: str = "\n".join(f"{k}: {v}" for agent in agents_response for k, v in agent.items())
            enriched_prompt: str = f"""Você é um agente gerente que tem sob sua responsabilidade alguns agentes especializados.

Você solicitou a chamada de alguns agentes. 
Os seguintes agentes responderam individualmente:
{combined}

Seu papel é responder ao usuário com base nas respostas dos agente. 
Para isso gere uma única resposta unificada e natural para o usuário final."""

            return enriched_prompt

        except Exception as e:
            print(f"[ERROR] - Falha ao gerar o prompt final do Manager: {e}")
            return ""

    def __update_history(self, user_input: str, agent_response: str, called_delegated_agents: bool, response_delegated_agents: List[dict], score: int | None = None) -> None:
        try:
            actual_conversation = {
                "id": str(uuid.uuid4()),
                "timestamp": datetime.now().strftime("%Y-%m-%dT%H:%M:%S"),
                "interaction": {
                    "user": user_input,
                    "agent": []
                },
                "type_agent": "manager",
                "score": score,
                "called_agents": called_delegated_agents
            }

            if called_delegated_agents:
                for agent in response_delegated_agents:
                    for agente_name, response in agent.items():  # -> Esse for é sempre fixo em 1 item.
                        actual_conversation["interaction"][agente_name] = [response]

            actual_conversation["interaction"]['agent'] = [agent_response]

            self.history.append(actual_conversation)
            self.history = self.history[-self.MAX_HISTORY:]  # -> Mantendo apenas os N itens no histórico.
            self.storage.save_history(self.agent_name, actual_conversation)
        except Exception as e:
            print(f'[ERROR] - Ocorreu um erro duração a atualização do histórico: {e}')



from typing import List, Optional, Union, Callable, Dict, Any
from google.genai import types
from tyr_agent.mixins.gemini_file_mixins import GeminiFileMixin
from tyr_agent.core.ai_config import configure_gemini
import json
import inspect


class GeminiModel(GeminiFileMixin):
    def __init__(self, model_name: str, temperature: Union[int, float] = 0.4, max_tokens: int = 600, api_key: Optional[str] = None):
        self.client = configure_gemini(api_key)

        self.model_name = model_name

        self.temperature = temperature
        self.max_tokens = max_tokens

    def generate(self, prompt_build: str, user_input: str, files: Optional[List[dict]], history: Optional[List[dict]], use_history: bool) -> str:
        messages = self.__create_messages(user_input, files, history, use_history)

        response = self.client.models.generate_content(
            model=self.model_name,
            contents=messages,
            config=types.GenerateContentConfig(
                system_instruction=prompt_build,
                max_output_tokens=self.max_tokens,
                temperature=self.temperature,
            )
        )

        return response.text.strip()

    async def async_generate(self, prompt_build: str, user_input: str, files: Optional[List[dict]], history: Optional[List[dict]], use_history: bool) -> str:
        messages = self.__create_messages(user_input, files, history, use_history)

        final_response: str = ""
        for chunk in self.client.models.generate_content_stream(
                model=self.model_name,
                contents=messages,
                config=types.GenerateContentConfig(
                    system_instruction=prompt_build,
                    max_output_tokens=self.max_tokens,
                    temperature=self.temperature,
                )
            ):
            final_response += chunk.text

        return final_response.strip()

    async def generate_with_functions(self, prompt_build: str, user_input: str, files: Optional[List[dict]], history: Optional[List[dict]], use_history: bool, functions: Optional[List[Callable]], final_prompt: Optional[str]):
        messages = self.__create_messages(user_input, files, history, use_history)

        response = self.client.models.generate_content(
            model=self.model_name,
            contents=messages,
            config=types.GenerateContentConfig(
                system_instruction=prompt_build,
                max_output_tokens=self.max_tokens,
                temperature=self.temperature,
                tools=functions if functions else None,
                automatic_function_calling=types.AutomaticFunctionCallingConfig(disable=True)
            ),
        )

        # Pegando as funções chamadas pelo modelo:
        calls = response.function_calls

        # Validando se teve alguma chamada de função:
        if not calls:
            return response.text.strip()  # Nenhuma função chamada, retorna direto

        tool_content = await self.__execute_functions(calls, functions)

        # Parte 5 - Segunda chamada: modelo continua raciocínio com base na resposta da função
        final_response = self.client.models.generate_content(
            model=self.model_name,
            contents=[
                *messages,
                response.candidates[0].content,  # chamada da função pelo modelo
                tool_content  # resposta da função
            ],
            config=types.GenerateContentConfig(
                system_instruction=final_prompt if final_prompt is not None else prompt_build,
                max_output_tokens=self.max_tokens,
                temperature=self.temperature
            ),
        )

        return final_response.text.strip()

    def __create_messages(self, user_input: str, files: Optional[List[dict]], history: Optional[List[dict]], use_history: bool) -> List[Any]:
        messages = self.__build_messages(user_input, history, use_history)

        if files:
            files_formated = [self.convert_item_to_gemini_model(item["file"], item["file_name"]) for item in files]
            files_valid = [file for file in files_formated if file]

            # Adicionando os arquivos identificados dentro do parts da pergunta atual do usuário:
            if files_valid:
                messages[-1].parts.extend(files_valid[:10])

        if not messages:
            raise Exception("[ERROR] - Erro ao gerar o prompt do GEMINI.")

        return messages

    def __build_messages(self, user_input: str, history: Optional[List[dict]], use_history: bool):
        messages: List = []

        if history and use_history:
            for interaction in history:
                user_text = interaction["interaction"]["user"]

                messages.append(
                    types.Content(
                        role="user",
                        parts=[types.Part.from_text(text=user_text)]
                    )
                )

                for agent_text in interaction["interaction"]["agent"]:
                    messages.append(
                        types.Content(
                            role="model",
                            parts=[types.Part.from_text(text=agent_text)]
                        )
                    )

        messages.append(
            types.Content(
                role="user",
                parts=[types.Part.from_text(text=user_input)]
            )
        )

        return messages

    async def __execute_functions(self, calls, functions: List[Callable]):
        # Parte 1: Criando um dicionário com o nome das funções e as funções:
        dict_functions: Dict[str, Callable] = {fn.__name__: fn for fn in functions}

        # Parte 2:
        tool_parts: list = []
        for call in calls:
            fn = dict_functions.get(call.name)
            if fn is None:
                raise Exception(f"[ERROR] - Função '{call.name}' não encontrada.")

            try:
                if inspect.iscoroutinefunction(fn):
                    result = await fn(**call.args)
                else:
                    result = fn(**call.args)
            except Exception as e:
                result = {"error": f"Ocorreu um erro durante a execução da função: {str(e)}"}

            part = types.Part.from_function_response(
                name=call.name,
                response={"result": result}
            )
            tool_parts.append(part)

        # Parte 3 - Cria o conteúdo do resultado das funções:
        tool_content = types.Content(role="tool", parts=tool_parts)

        return tool_content



import base64


def image_to_base64(caminho_imagem: str) -> str:
    """
    Converte uma imagem para uma string em base64.

    Parâmetros:
        caminho_imagem (str): Caminho do arquivo da imagem (ex: 'imagem.jpg').

    Retorna:
        str: String base64 da imagem.
    """
    try:
        with open(caminho_imagem, "rb") as imagem:
            imagem_bytes = imagem.read()
            imagem_base64 = base64.b64encode(imagem_bytes).decode('utf-8')
            return imagem_base64
    except FileNotFoundError:
        print(f"Arquivo não encontrado: {caminho_imagem}")
        return ""
    except Exception as e:
        print(f"Erro ao converter imagem: {e}")
        return ""



def configure_gemini(api_key: str | None = None):
    import os
    from dotenv import load_dotenv
    from google import genai

    load_dotenv()
    key = api_key or os.getenv("GEMINI_KEY")
    if not key:
        raise EnvironmentError("API key não definida.")
    return genai.Client(api_key=key)


def configure_gpt(api_key: str | None = None):
    import os
    from openai import OpenAI
    from dotenv import load_dotenv

    load_dotenv()
    key = api_key or os.getenv("OPENAI_API_KEY")
    if not key:
        raise EnvironmentError("OPENAI_API_KEY não definida.")
    return OpenAI(api_key=key)



from typing import TypedDict, List, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from tyr_agent import SimpleAgent, ComplexAgent


class ManagerCallAgent(TypedDict):
    agent_to_call: str
    agent_message: str


class ManagerCallManyAgents(TypedDict):
    call_agents: bool
    agents_to_call: List[ManagerCallAgent]


class AgentCallInfo(TypedDict):
    agent: "SimpleAgent | ComplexAgent"
    message: str


class AgentInteraction(TypedDict):
    user: str
    agent: List[str]


class AgentHistory(TypedDict):
    timestamp: Optional[str]
    interaction: AgentInteraction
    score: Optional[Union[int, float]]
    type_agent: str



from openai import OpenAI
from typing import Optional, Union, Callable, List, Dict, Any
from openai.types.responses import ResponseTextConfigParam
from tyr_agent.core.ai_config import configure_gpt
from tyr_agent.mixins.gpt_file_mixins import GPTFileMixin
from tyr_agent.utils.gpt_function_format_utils import to_openai_tool
import json
import inspect


class GPTModel(GPTFileMixin):
    def __init__(self, model_name: str, temperature: Union[int, float] = 0.4, max_tokens: int = 600, effort: str = "medium", response_template: Optional[ResponseTextConfigParam] = None, api_key: Optional[str] = None):
        self.client: OpenAI = configure_gpt(api_key)

        if model_name == "economy":
            self.model_name = "gpt-3.5-turbo"
        elif model_name == "quality":
            self.model_name = "gpt-5.1"
        else:
            self.model_name = model_name

        if effort not in ["low", "medium", "high"]:
            self.effort = "medium"

        self.temperature = temperature
        self.max_tokens = max_tokens
        self.effort = effort
        self.response_template = response_template

    def generate(self, prompt_build: str, user_input: str, files: Optional[List[dict]], history: Optional[List[dict]], use_history: bool) -> str:
        messages = self.__create_messages(prompt_build, user_input, files, history, use_history)

        response = self.client.responses.create(
            model=self.model_name,
            reasoning={"effort": self.effort},
            max_output_tokens=self.max_tokens,
            input=messages,
            text=self.response_template
        )

        return response.output_text


    async def async_generate(self, prompt_build: str, user_input: str, files: Optional[List[dict]], history: Optional[List[dict]], use_history: bool) -> str:
        pass

    async def generate_with_functions(self, prompt_build: str, user_input: str, files: Optional[List[dict]], history: Optional[List[dict]], use_history: bool, functions: Optional[List[Callable]], final_prompt: Optional[str]):
        messages = self.__create_messages(prompt_build, user_input, files, history, use_history)

        # Criando um array com as funções no formato que o GPT precisa:
        tools = []
        for f in functions:
            tools.append(to_openai_tool(f, True))

        response = self.client.responses.create(
            model=self.model_name,
            reasoning={"effort": self.effort},
            max_output_tokens=self.max_tokens,
            input=messages,
            tools=tools if tools else None,
            text=self.response_template
        )

        # Pegando as funções chamadas pelo modelo:
        calls = response.output

        # Validando se teve alguma chamada de função:
        if not calls or not any(call.type == "function_call" for call in calls):
            return response.output_text  # Nenhuma função chamada, retorna direto

        new_messages = await self.__execute_functions(calls, messages, functions)

        # Alterando o prompt "system" das mensagens pra o prompt especial definido na inicialização do agente:
        if final_prompt:
            new_messages[0]["content"] = final_prompt

        # Parte 5 - Segunda chamada: modelo continua raciocínio com base na resposta da função
        response_answer_functions = self.client.responses.create(
            model=self.model_name,
            reasoning={"effort": self.effort},
            max_output_tokens=self.max_tokens,
            input=new_messages,
            text=self.response_template
        )

        return response_answer_functions.output_text

    def __create_messages(self, prompt_build, user_input: str, files: Optional[List[dict]], history: Optional[List[dict]], use_history: bool) -> List[Any]:
        messages = self.__build_messages(prompt_build, user_input, history, use_history)

        if files:
            files_formated = [self.convert_item_to_gpt_model(item["file"], item["file_name"]) for item in files]
            files_valid = []
            for file in files_formated:
                if file is None:
                    continue

                if file[0].startswith("data:application/pdf"):
                    files_valid.append({"type": "input_file", "filename": file[1], "file_data": file[0]})
                    continue

                files_valid.append({"type": "input_image", "image_url": file[0]})

            # Adicionando os arquivos identificados dentro da pergunta atual do usuário:
            if files_valid:
                messages[-1] = {
                    "role": "user",
                    "content": [
                        {"type": "input_text", "text": user_input},
                        *files_valid[:10]
                    ]
                }

        if not messages:
            raise Exception("[ERROR] - Erro ao gerar o prompt do GPT.")

        return messages

    def __build_messages(self, prompt_build: str, user_input: str, history: Optional[List[dict]], use_history: bool) -> List[Any]:
        messages: List[dict] = [{"role": "system", "content": prompt_build}]

        if history and use_history:
            for interaction in history:
                user_text = interaction["interaction"]["user"]

                messages.append({"role": "user", "content": user_text})

                for agent_text in interaction["interaction"]["agent"]:
                    messages.append({"role": "assistant", "content": agent_text})

        messages.append({"role": "user", "content": user_input})

        return messages

    async def __execute_functions(self, calls, messages, functions: List[Callable]):
        # Parte 1: Criando um dicionário com o nome das funções e as funções:
        dict_functions: Dict[str, Callable] = {fn.__name__: fn for fn in functions}

        # Parte 2: Adicionando a mensagem do GPT solicitando a execução das funções no histórico:
        messages += calls

        # Parte 3: Executando as funções solicitadas pelo GPT:
        for call in calls:
            if call.type == "function_call":
                fn = dict_functions.get(call.name)
                if fn is None:
                    raise Exception(f"[ERROR] - Função '{call.name}' não encontrada.")
                try:
                    args = json.loads(call.arguments)
                    if inspect.iscoroutinefunction(fn):
                        result = await fn(**args)
                    else:
                        result = fn(**args)
                except Exception as e:
                    result = {"error": f"Ocorreu um erro durante a execução da função: {str(e)}"}

                # Parte 4: Adicionando a resposta da função executada ao histórico de mensagens:
                messages.append({
                    "type": "function_call_output",
                    "call_id": call.call_id,
                    "output": json.dumps({call.name: result})
                })

        return messages



import json
import os
from typing import List


class InteractionHistory:
    def __init__(self, filename: str = "conversation_history.json"):
        self.filename = filename
        if not os.path.exists(self.filename):
            with open(self.filename, "w", encoding="utf-8") as f:
                json.dump({}, f, ensure_ascii=False)

    def save_history(self, agent_name: str, history: dict) -> None:
        try:
            data = self.load_all()

            if data.get(agent_name, False):
                data[agent_name].append(history)
            else:
                data[agent_name] = [history]

            with open(self.filename, "w", encoding="utf-8") as f:
                json.dump(data, f, indent=2, ensure_ascii=False)
        except Exception as e:
            print(f"[ERROR] - Erro ao salvar histórico: {e}")

    def load_history(self, agent_name: str) -> List[dict]:
        data = self.load_all()
        return data.get(agent_name, [])

    def load_all(self) -> dict:
        try:
            with open(self.filename, "r", encoding="utf-8") as f:
                return json.load(f)
        except FileNotFoundError:
            return {}

    def clear_history(self) -> None:
        try:
            with open(self.filename, "w", encoding="utf-8") as f:
                json.dump({}, f, indent=2, ensure_ascii=False)
        except Exception as e:
            print(f"[ERROR] - Erro ao limpar o histórico.")

    def update_score(self, agent_name: str, interaction_id: str, score: float) -> bool:
        try:
            if not isinstance(score, (int, float)) or not (0 <= score <= 5):
                raise ValueError("Score deve ser um número entre 0 e 5.")

            data = self.load_all()

            if data.get(agent_name, False):
                for interaction in data[agent_name]:
                    if interaction["id"] == interaction_id:
                        interaction["score"] = score

                        with open(self.filename, "w", encoding="utf-8") as f:
                            json.dump(data, f, indent=2, ensure_ascii=False)

                        return True
            else:
                return False
        except Exception as e:
            print(f"[ERROR] - Erro ao atualizar o score: {e}")
            return False

    def delete_history(self, agent_name: str, interaction_id: str) -> bool:
        try:
            data = self.load_all()

            if data.get(agent_name, False):
                data[agent_name] = list(filter(lambda x: x["id"] != interaction_id, data[agent_name]))

                with open(self.filename, "w", encoding="utf-8") as f:
                    json.dump(data, f, indent=2, ensure_ascii=False)

                return True
            else:
                return False
        except Exception as e:
            print(f"[ERROR] - Erro ao excluir interação: {e}")
            return False



from .core.agent import SimpleAgent, ComplexAgent, ManagerAgent
from .core.ai_config import configure_gemini, configure_gpt
from .storage.interaction_history import InteractionHistory
from .utils.image_utils import image_to_base64
from .mixins.gemini_file_mixins import GeminiFileMixin
from .mixins.gpt_file_mixins import GPTFileMixin
from .models.gemini_model import GeminiModel
from .models.gpt_model import GPTModel
from .entities.entities import AgentInteraction, AgentHistory

__all__ = [
    "SimpleAgent",
    "ComplexAgent",
    "ManagerAgent",
    "configure_gemini",
    "configure_gpt",
    "InteractionHistory",
    "GeminiModel",
    "GPTModel",
    "AgentInteraction",
    "AgentHistory"
]


from typing import Optional, Union
from io import BytesIO
from google.genai import types


class GeminiFileMixin:
    def convert_item_to_gemini_model(self, file: Union[str, BytesIO], file_name: str) -> Optional[types.Part]:
        """
        Converte um path, base64 ou BytesIO para o formato ideal para ser enviado na requisição do Gemini.
        :param file: Arquivo a ser enviado, podendo ser path, base64 ou BytesIO.
        :param file_name: Nome do arquivo a ser enviado.
        :return: Retorna um dicionário no formato ideal para ser enviado via Gemini.
        """
        try:
            bytes_file = self.__get_file_bytes(file)
            if not bytes_file:
                return None

            mime_type = self.__detect_mime_type(file_name)

            if mime_type == "application/octet-stream":
                raise ValueError(f"Tipo MIME não suportado ou não reconhecido para o arquivo '{file_name}'.")

            return types.Part.from_bytes(data=bytes_file, mime_type=mime_type)
        except Exception as e:
            return None

    def __get_file_bytes(self, file: Union[str, BytesIO]) -> Optional[bytes]:
        """
        Pega os bytes de um arquivo, seja via path, base64 ou BytesIO.
        :param file: Arquivo que terá seus bytes extraídos.
        :return: Retorna os bytes do arquivo.
        """
        if isinstance(file, str):
            bytes_file = self.__convert_base64_to_bytes(file)
            if bytes_file is None:  # Não é base64, então é path.
                try:
                    with open(file, "rb") as f:
                        bytes_file = f.read()
                except Exception:
                    return None
            return bytes_file
        elif isinstance(file, BytesIO):
            return file.read()
        return None

    def __detect_mime_type(self, file_name: str) -> str:
        """
        Detecta o tipo do arquivo informado baseado no nome dele e retorno mime type dele.
        :param file_name: Nome do arquivo a ser analisado.
        :return: Retorna o tipo mime do arquivo recebido.
        """
        import mimetypes
        from pathlib import Path

        mime_map = {
            ".pdf": "application/pdf",
            ".csv": "text/csv",
            ".txt": "text/plain",
            ".jpg": "image/jpeg",
            ".jpeg": "image/jpeg",
            ".png": "image/png",
            ".json": "application/json",
        }

        ext = Path(file_name).suffix.lower()
        if ext in mime_map:
            return mime_map[ext]

        mime_type, _ = mimetypes.guess_type(file_name)
        return mime_type or "application/octet-stream"

    def __convert_base64_to_bytes(self, string_base64: str) -> Optional[bytes]:
        """
        Tenta decodificar um conteúdo base64 (com ou sem header).
        :param string_base64: String a ser validada.
        :return: Retorna os bytes ou levanta None se não for válido.
        """
        import base64

        try:
            if string_base64.startswith("data:"):
                header, b64data = string_base64.split(",", 1)
            else:
                b64data = string_base64

            # Validando o tamanho e padding: (b64 precisa ser múltiplo de 4)
            if len(b64data.strip()) % 4 != 0:
                b64data += '=' * (4 - len(b64data.strip()) % 4)

            # Decodificando a string base64 (b64data) para bytes e Validando se a string é um base64 válido e bem formado:
            return base64.b64decode(b64data, validate=True)
        except Exception:
            # print(f'[ERROR] - __detect_base64: {type(e).__name__}: {e}')
            return None



from typing import Optional, Union, Tuple
from io import BytesIO


class GPTFileMixin:
    SUPPORTED_TYPES = {
        "image/jpeg",
        "image/png",
        "image/webp",
        "application/pdf"
    }

    def convert_item_to_gpt_model(self, file: Union[str, bytes, BytesIO], file_name: str) -> Optional[Tuple[str, str]]:
        """
        Converte um arquivo (path, bytes ou BytesIO) para base64 com prefixo data URL (data:image/png;base64,...).
        Suporta apenas imagens.
        :param file: Caminho para o arquivo, bytes ou BytesIO.
        :param file_name: Nome original do arquivo (usado para detectar o mime type).
        :return: Tupla de Strings onde 0: base64 formatada como data URL; 1: Nome do arquivo.
        :raises: ValueError para tipos de arquivo não suportados ou erros de leitura.
        """
        import base64

        # Detecta o mime type com base no nome
        mime_type = self.__detect_mime_type(file_name)

        if mime_type not in self.SUPPORTED_TYPES:
            return None

        # Obtendo os bytes do arquivo:
        bytes_file = self.__get_file_bytes(file)
        if not bytes_file:
            return None

        # Codificando em base64:
        b64_string = base64.b64encode(bytes_file).decode()

        # Retorna no formato data URL
        return f"data:{mime_type};base64,{b64_string}", file_name

    def __detect_mime_type(self, file_name: str) -> str:
        """
        Detecta o tipo do arquivo informado baseado no nome dele e retorno mime type dele.
        :param file_name: Nome do arquivo a ser analisado.
        :return: Retorna o tipo mime do arquivo recebido.
        """
        import mimetypes
        from pathlib import Path

        mime_map = {
            ".jpg": "image/jpeg",
            ".jpeg": "image/jpeg",
            ".png": "image/png",
            ".webp": "image/webp",
            ".pdf": "application/pdf"
        }

        ext = Path(file_name).suffix.lower()
        if ext in mime_map:
            return mime_map[ext]

        mime_type, _ = mimetypes.guess_type(file_name)
        return mime_type or "application/octet-stream"

    def __get_file_bytes(self, file: Union[str, BytesIO]) -> Optional[bytes]:
        """
        Pega os bytes de um arquivo, seja via path ou BytesIO.
        :param file: Arquivo que terá seus bytes extraídos.
        :return: Retorna os bytes do arquivo.
        """
        if isinstance(file, str):
            try:
                with open(file, "rb") as f:
                    return f.read()
            except Exception:
                return None
        elif isinstance(file, BytesIO):
            return file.read()
        elif isinstance(file, bytes):
            return file
        return None
